# -*- coding: utf-8 -*-
#
#  Copyright 2018-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from ctypes import c_char, c_double, c_short, c_long, create_string_buffer, POINTER, Structure, cdll, byref
from distutils.util import get_platform
from io import StringIO, TextIOWrapper
from logging import warning
from os import name
from pathlib import Path
from re import split
from sys import prefix, exec_prefix
from traceback import format_exc
from typing import List, Optional
from warnings import warn
from ._CGRrw import CGRRead, common_isotopes, parse_error
from ..containers import MoleculeContainer


class INCHIRead(CGRRead):
    """
    INCHI separated per lines files reader. works similar to opened file object. support `with` context manager.
    on initialization accept opened in text mode file, string path to file,
    pathlib.Path object or another buffered reader object.
    line should be start with INCHI string and
    optionally continues with space/tab separated list of key:value [or key=value] data if header=None.
        example:
            InChI=1S/C2H5/c1-2/h1H2,2H3/q+1 id:123 key=value
    if header=True then first line of file should be space/tab separated list of keys including INCHI column key.
        example:
            ignored_inchi_key key1 key2
            InChI=1S/C2H5/c1-2/h1H2,2H3/q+1 1 2
    also possible to pass list of keys (without inchi_pseudo_key) for mapping space/tab separated list
    of INCHI and values: header=['key1', 'key2'] # order depended
    """
    def __init__(self, file, header=None, ignore_stereo=False, **kwargs):
        """
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `CGRtoolsParserLog`.
        :param calc_cis_trans: Calculate cis/trans marks from 2d coordinates.
        :param ignore_stereo: Ignore stereo data.
        """
        if isinstance(file, str):
            self._file = open(file)
            self.__is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open()
            self.__is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO)):
            self._file = file
            self.__is_buffer = True
        else:
            raise TypeError('invalid file. TextIOWrapper, StringIO subclasses possible')
        super().__init__(**kwargs)
        self.__file = iter(self._file.readline, '')

        if header is True:
            self.__header = next(self.__file).split()[1:]
        elif header:
            if not isinstance(header, (list, tuple)) or not all(isinstance(x, str) for x in header):
                raise TypeError('expected list (tuple) of strings')
            self.__header = header
        else:
            self.__header = None

        self.__ignore_stereo = ignore_stereo
        self._data = self.__data()

    def __data(self):
        file = self._file
        parse = self.parse
        seekable = file.seekable()
        pos = file.tell() if seekable else None
        for n, line in enumerate(self.__file):
            x = parse(line)
            if x is None:
                yield parse_error(n, pos, self._format_log())
                if seekable:
                    pos = file.tell()
            else:
                yield x

    @classmethod
    def create_parser(cls, *args, **kwargs):
        """
        Create INCHI parser function configured same as INCHIRead object
        """
        obj = object.__new__(cls)
        obj._INCHIRead__header = None
        obj._INCHIRead__ignore_stereo = False
        super(INCHIRead, obj).__init__(*args, **kwargs)
        return obj.parse

    def close(self, force=False):
        """
        close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self.__is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def read(self) -> List[Optional[MoleculeContainer]]:
        """
        parse whole file

        :return: list of parsed molecules
        """
        return list(iter(self))

    def __iter__(self):
        return (x for x in self._data if not isinstance(x, parse_error))

    def __next__(self):
        return next(iter(self))

    def parse(self, inchi: str) -> Optional[MoleculeContainer]:
        """
        convert INCHI string into MoleculeContainer object. string should be start with INCHI and
        optionally continues with space/tab separated list of key:value [or key=value] data.
        """
        self._flush_log()
        inchi, *data = inchi.split()
        if not inchi:
            self._info('empty inchi string')
            return
        elif self.__header is None:
            meta = {}
            for x in data:
                try:
                    k, v = split('[=:]', x, 1)
                    meta[k] = v
                except ValueError:
                    self._info(f'invalid metadata entry: {x}')
        else:
            meta = dict(zip(self.__header, data))

        try:
            record = self.__parse_inchi(inchi)
        except ValueError:
            self._info(f'string: {inchi}\nconsist errors:\n{format_exc()}')
            return

        record['meta'] = meta
        try:
            container = self._convert_structure(record)
        except ValueError:
            self._info(f'record consist errors:\n{format_exc()}')
        else:
            if self._store_log:
                log = self._format_log()
                if log:
                    container.meta['CGRtoolsParserLog'] = log
            return container

    @staticmethod
    def __parse_inchi(string):
        structure = INCHIStructure()
        if lib.GetStructFromINCHI(byref(InputINCHI(string)), byref(structure)):
            lib.FreeStructFromINCHI(byref(structure))
            raise ValueError('invalid INCHI')

        atoms, bonds = [], []
        seen = set()
        for n in range(structure.num_atoms):
            seen.add(n)
            atom = structure.atom[n]
            element = atom.elname.decode()

            isotope = atom.isotopic_mass
            if isotope in (0, 10000):
                isotope = None
            elif isotope > 10000:
                isotope = isotope - 10000 + common_isotopes[element]

            atoms.append({'element': element, 'charge': int.from_bytes(atom.charge, byteorder='big', signed=True),
                          'mapping': 0, 'x': atom.x, 'y': atom.y, 'z': atom.z, 'isotope': isotope,
                          'is_radical': bool(int.from_bytes(atom.radical, byteorder='big'))})

            for k in range(atom.num_bonds):
                m = atom.neighbor[k]
                if m in seen:
                    continue
                order = atom.bond_type[k]
                if order:
                    bonds.append((n, m, order))

        lib.FreeStructFromINCHI(byref(structure))
        return {'atoms': atoms, 'bonds': bonds}


class InputINCHI(Structure):
    def __init__(self, string, options=None):
        if options is None:
            options = create_string_buffer(1)
        else:
            options = create_string_buffer(' '.join(f'{opt_flag}{x}' for x in options).encode())
        super().__init__(create_string_buffer(string.encode()), options)

    _fields_ = [('szInChI', POINTER(c_char)),  # InChI ASCII string to be converted to a strucure
                ('szOptions', POINTER(c_char))  # InChI options: space-delimited; each is preceded
                                                # by '/' or '-' depending on OS and compiler
                ]


class Atom(Structure):
    _fields_ = [('x', c_double), ('y', c_double), ('z', c_double),  # atom coordinates
                ('neighbor', c_short * 20),  # adjacency list: ordering numbers of the adjacent atoms, >= 0
                ('bond_type', c_char * 20),  # inchi_BondType
                ('bond_stereo', c_char * 20),  # inchi_BondStereo2D; negative if the sharp end points to opposite atom
                ('elname', c_char * 6),  # zero-terminated chemical element name: "H", "Si", etc.
                ('num_bonds', c_short),  # number of neighbors, bond types and bond stereo in the adjacency list
                ('num_iso_H', c_char * 4),  # implicit hydrogen atoms
                                            # [0]: number of implicit non-isotopic H (exception: num_iso_H[0]=-1 means
                                            # INCHI adds implicit H automatically),
                                            # [1]: number of implicit isotopic 1H (protium),
                                            # [2]: number of implicit 2H (deuterium),
                                            # [3]: number of implicit 3H (tritium)
                ('isotopic_mass', c_short),  # 0 => non-isotopic; isotopic mass or 10000 + mass - average atomic mass
                ('radical', c_char),  # inchi_Radical,
                ('charge', c_char)  # positive or negative; 0 => no charge
                ]


class Stereo0D(Structure):
    _fields_ = [('neighbor', c_short * 4),  # 4 atoms always
                ('central_atom', c_short),  # central tetrahedral atom or a central atom of allene; otherwise NO_ATOM
                ('type', c_char),  # inchi_StereoType0D
                ('parity', c_char)  # inchi_StereoParity0D: may be a combination of two parities:
                                    # ParityOfConnected | (ParityOfDisconnected << 3), see Note above
                ]


class INCHIStructure(Structure):
    _fields_ = [('atom', POINTER(Atom)),  # array of num_atoms elements
                ('stereo0D', POINTER(Stereo0D)),  # array of num_stereo0D 0D stereo elements or NULL
                ('num_atoms', c_short),  # number of atoms in the structure
                ('num_stereo0D', c_short),  # number of 0D stereo elements
                ('szMessage', POINTER(c_char)),  # Error/warning ASCII message
                ('szLog', POINTER(c_char)),  # log-file ASCII string, contains a human-readable list
                                             # of recognized options and possibly an Error/warn message
                ('WarningFlags', (c_long * 2) * 2)
                ]


class INCHIread:
    def __init__(self, *args, **kwargs):
        warn('INCHIread deprecated. Use INCHIRead instead', DeprecationWarning)
        warning('INCHIread deprecated. Use INCHIRead instead')
        self.__obj = INCHIRead(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __iter__(self):
        return iter(self.__obj)

    def __next__(self):
        return next(self.__obj)

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, _type, value, traceback):
        return self.__obj.__exit__(_type, value, traceback)


def getsitepackages():
    """returns a list containing all global site-packages directories. stolen and modified site.py function
    """
    sitepackages = []
    for pr in {prefix, exec_prefix}:
        pr = Path(pr)
        if name == 'posix':
            sitepackages.append(pr / 'local/lib')
        else:
            sitepackages.append(pr)
        sitepackages.append(pr / 'lib')
    return sitepackages


platform = get_platform()
if platform in ('linux-x86_64', 'win-amd64'):
    if platform == 'win-amd64':
        opt_flag = '/'
        libname = 'libinchi.dll'
    else:
        opt_flag = '-'
        libname = 'libinchi.so'

    for site in getsitepackages():
        lib_path = site / libname
        if lib_path.exists():
            lib = cdll.LoadLibrary(str(lib_path))
            __all__ = ['INCHIRead', 'INCHIread']
            break
    else:
        warn('broken package installation. libinchi not found', ImportWarning)
        __all__ = []
        del INCHIRead, INCHIread
else:
    warn('unsupported platform', ImportWarning)
    __all__ = []
    del INCHIRead, INCHIread



# -*- coding: utf-8 -*-
#
#  Copyright 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2020 Dinar Batyrshin <batyrshin-dinar@mail.ru>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from math import acos, sqrt
from collections import defaultdict


def plane_normal(nmx, nmy, nmz, nox, noy, noz):
    # return normal to plane of two vectors nm and no
    # m <--- n
    #         \
    #          v
    #          o
    return nmy * noz - nmz * noy, nox * nmz - nmx * noz, nmx * noy - nmy * nox


def unit_vector(nmx, nmy, nmz):
    nmd = sqrt(nmx ** 2 + nmy ** 2 + nmz ** 2)
    return nmx / nmd, nmy / nmd, nmz / nmd


def get_angle(nx, ny, nz, mx, my, mz):
    ch = (nx * mx + ny * my + nz * mz) ** 2
    zn = (nx ** 2 + ny ** 2 + nz ** 2) * (mx ** 2 + my ** 2 + mz ** 2)
    if ch < .0001:
        return 1.
    elif zn < .0001:
        return .0
    else:
        return sqrt(1 - ch / zn)


def vector_normal(nmx, nmy, nmz):
    # return normal to vector nm
    if not -.0001 < nmx < .0001:
        return (- nmy - nmz) / nmx, 1, 1
    elif not -.0001 < nmy < .0001:
        return 1, (- nmx - nmz) / nmy, 1
    else:
        return 1, 1, (- nmx - nmy) / nmz


class JupyterWidget:
    def __init__(self, xml, width, height):
        self.xml = xml
        self.width = width
        self.height = height

    def _repr_html_(self):
        return ("<script type='text/javascript' src='https://www.x3dom.org/download/x3dom.js'></script>"
                "<link rel='stylesheet' type='text/css' href='https://www.x3dom.org/download/x3dom.css'>"
                f'<div style="width: {self.width}; height: {self.height}">{self.xml}</div>')

    def __html__(self):
        return self._repr_html_()


class X3dom:
    __slots__ = ()

    def depict3d(self, index: int = 0) -> str:
        """Get X3DOM XML string.

        :param index: index of conformer
        """
        xyz = self._conformers[index]
        mx = sum(x for x, _, _ in xyz.values()) / len(xyz)
        my = sum(y for _, y, _ in xyz.values()) / len(xyz)
        mz = sum(z for _, _, z in xyz.values()) / len(xyz)
        xyz = {n: (x - mx, y - my, z - mz) for n, (x, y, z) in xyz.items()}
        atoms = self.__render_atoms(xyz)
        bonds = self._render_3d_bonds(xyz)
        return f'<x3d width=100% height=100%>\n  <scene>\n{atoms}{bonds}  </scene>\n</x3d>'

    def view3d(self, index: int = 0, width='600px', height='400px'):
        """
        Jupyter widget for 3D visualization.

        :param index: index of conformer
        :param width: widget width
        :param height: widget height
        """
        return JupyterWidget(self.depict3d(index), width, height)

    def __render_atoms(self, xyz):
        config = self._render_config

        font = config['font_size']
        carbon = config['carbon']
        radius = config['atom_radius']
        colors = config['atoms_colors']
        mapping_color = config['mapping_color']

        if radius < 0:
            multiplier = -radius
            radius = 0
        elif not radius:
            multiplier = .2

        atoms = []
        if carbon:
            for n, a in self._atoms.items():
                r = radius or a.atomic_radius * multiplier
                fr = r * 0.71
                atoms.append(f"    <transform translation='{' '.join(format(x, '.2f') for x in xyz[n])}'>\n"
                             "      <billboard axisOfRotation='0 0 0'>\n        <group>\n"
                             f"          <transform translation='{fr:.2f} {fr:.2f} 0'>\n"
                             '            <shape>\n              <appearance>\n'
                             f"                <material diffuseColor='{mapping_color}'/>\n"
                             f"              </appearance>\n              <text string='{a.atomic_symbol}'>\n"
                             f"                <fontstyle family='sans' size='{font:.2f}' justify='first'/>\n"
                             "              </text>\n            </shape>\n          </transform>\n"
                             "          <shape>\n            <appearance>\n"
                             f"              <material diffuseColor='{colors[a.atomic_number - 1]}'/>\n"
                             f"            </appearance>\n            <sphere radius='{r:.2f}'/>\n"
                             "          </shape>\n        </group>\n      </billboard>\n    </transform>\n")
        else:
            for n, a in self._atoms.items():
                r = radius or a.atomic_radius * multiplier
                atoms.append(f"    <transform translation='{' '.join(format(x, '.2f') for x in xyz[n])}'>\n"
                             "      <shape>\n        <appearance>\n"
                             f"          <material diffuseColor='{colors[a.atomic_number - 1]}'/>\n"
                             f"        </appearance>\n        <sphere radius='{r:.2f}'/>\n"
                             "      </shape>\n    </transform>\n")
        return ''.join(atoms)

    def _render_3d_dashes(self, nx, ny, nz, nmx, nmy, nmz, nm_ln, r_angle=None, cgr_color=''):
        config = self._render_config

        bond_radius = config['bond_radius']
        if cgr_color:
            bond_color = config[cgr_color]
        else:
            bond_color = config['bond_color']

        if r_angle is None:
            dash1, dash2 = config['aromatic_dashes']
            r_angle = acos(nmy / nm_ln)
        else:
            dash1, dash2 = config['dashes']

        xml = []
        dashes_sum = dash1 + dash2
        if dashes_sum < .0001:
            raise ValueError('Dashes should be nonzero')

        d = dashes_sum / nm_ln
        dx, dy, dz = nmx * d, nmy * d, nmz * d
        b = int((nm_ln - dash1) // dashes_sum)
        t = (nm_ln - (b * dashes_sum)) / nm_ln
        nx, ny, nz = nx + nmx * t / 2, ny + nmy * t / 2, nz + nmz * t / 2
        for _ in range(b):
            xml.append(f"    <transform translation='{nx:.2f} {ny:.2f} {nz:.2f}' rotation='{nmz:.2f} 0 "
                       f"{-nmx:.2f} {r_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                       f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                       f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{dash1:.2f}'>\n"
                       "        </cylinder>\n      </shape>\n    </transform>\n")
            nx += dx
            ny += dy
            nz += dz
        xml.append(f"    <transform translation='{nx:.2f} {ny:.2f} {nz:.2f}' rotation='{nmz:.2f} 0 "
                   f"{-nmx:.2f} {r_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                   f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                   f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{dash1:.2f}'>\n"
                   "        </cylinder>\n      </shape>\n    </transform>\n")
        return xml

    def _render_3d_aromatic_bond(self, n_x, n_y, n_z, m_x, m_y, m_z, c_x, c_y, c_z):
        aromatic_space = self._render_config['aromatic_space']

        # n aligned xyz
        nc_x, nc_y, nc_z = c_x - n_x, c_y - n_y, c_z - n_z
        mc_x, mc_y, mc_z = c_x - m_x, c_y - m_y, c_z - m_z

        nc_ln = sqrt(nc_x ** 2 + nc_y ** 2 + nc_z ** 2)
        mc_ln = sqrt(mc_x ** 2 + mc_y ** 2 + mc_z ** 2)
        sin1 = get_angle(m_x - n_x, m_y - n_y, m_z - n_z, nc_x, nc_y, nc_z)
        sin2 = get_angle(n_x - m_x, n_y - m_y, n_z - m_z, mc_x, mc_y, mc_z)

        if sin1 < .0001 or sin2 < .0001 or nc_ln < .0001 or mc_ln < .0001:
            return
        else:
            coef1 = aromatic_space / (nc_ln * sin1)
            coef2 = aromatic_space / (mc_ln * sin2)
            return nc_x * coef1, nc_y * coef1, nc_z * coef1, mc_x * coef2, mc_y * coef2, mc_z * coef2


class X3domMolecule(X3dom):
    __slots__ = ()

    def _render_3d_bonds(self, xyz):
        bonds = self._bonds
        config = self._render_config

        bond_color = config['bond_color']
        bond_radius = config['bond_radius']
        double_space = config['double_space']
        triple_space = config['triple_space']
        r1 = triple_space * sqrt(3) / 3
        r2 = triple_space * sqrt(3) / 6

        xml = []
        lengths = {}
        doubles = {}
        half_triple = triple_space / 2
        for n, m, bond in self.bonds():
            order = bond.order
            nx, ny, nz = xyz[n]
            mx, my, mz = xyz[m]

            nmx, nmy, nmz = mx - nx, my - ny, mz - nz
            length = sqrt(nmx ** 2 + nmy ** 2 + nmz ** 2)
            if length < .001:
                continue

            rotation_angle = acos(nmy / length)
            lengths[(n, m)] = lengths[(m, n)] = (length, rotation_angle)
            x, y, z = nx + nmx / 2, ny + nmy / 2, nz + nmz / 2
            if order in (1, 4):
                xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                           f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                           f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                           f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                           "        </cylinder>\n      </shape>\n    </transform>\n")
            elif order == 2:
                if n in doubles:
                    # normal for plane n m o
                    norm_x, norm_y, norm_z = plane_normal(nmx, nmy, nmz, *doubles[n])
                elif m in doubles:
                    # normal for plane n m o
                    norm_x, norm_y, norm_z = plane_normal(nmx, nmy, nmz, *doubles[m])
                else:
                    third = next((x for x in bonds[n] if x != m), None)
                    if third:
                        ox, oy, oz = xyz[third]
                        nox, noy, noz = ox - nx, oy - ny, oz - nz
                    else:
                        third = next((x for x in bonds[m] if x != n), None)
                        if third:
                            ox, oy, oz = xyz[third]
                            nox, noy, noz = ox - nx, oy - ny, oz - nz
                        else:
                            nox, noy, noz = vector_normal(nmx, nmy, nmz)

                    # normal for plane n m o
                    normx, normy, normz = unit_vector(*plane_normal(nmx, nmy, nmz, nox, noy, noz))

                    # normal for plane n m normal
                    norm_x, norm_y, norm_z = plane_normal(nmx, nmy, nmz, normx, normy, normz)

                doubles[n] = doubles[m] = (norm_x, norm_y, norm_z)
                norm_dist = sqrt(norm_x ** 2 + norm_y ** 2 + norm_z ** 2)

                if norm_dist < .0001:
                    coef = double_space * 10000
                else:
                    coef = double_space / norm_dist

                dx, dy, dz = norm_x * coef, norm_y * coef, norm_z * coef
                xml.append(
                    f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                    f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                    f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                    f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                    "        </cylinder>\n      </shape>\n    </transform>\n")
                xml.append(
                    f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                    f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                    f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                    f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                    "        </cylinder>\n      </shape>\n    </transform>\n")
            elif order == 3:
                nox, noy, noz = vector_normal(nmx, nmy, nmz)

                # normal for plane n m o
                normx, normy, normz = unit_vector(*plane_normal(nmx, nmy, nmz, nox, noy, noz))
                vecrx, vecry, vecrz = normx * r1, normy * r1, normz * r1

                # normal for plane n m normal
                norm_x, norm_y, norm_z = unit_vector(*plane_normal(nmx, nmy, nmz, normx, normy, normz))
                vecx, vecy, vecz = norm_x * half_triple, norm_y * half_triple, norm_z * half_triple

                xml.append(f"    <transform translation='{x + vecrx:.2f} {y + vecry:.2f} {z + vecrz:.2f}'"
                           f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                           f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                           f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                           f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")

                xx, yy, zz = x - normx * r2, y - normy * r2, z - normz * r2
                xml.append(f"    <transform translation='{xx - vecx:.2f} {yy - vecy:.2f} {zz - vecz:.2f}'"
                           f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                           f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                           f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                           f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                xml.append(f"    <transform translation='{xx + vecx:.2f} {yy + vecy:.2f} {zz + vecz:.2f}'"
                           f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                           f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                           f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                           f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
            else:
                xml.extend(self._render_3d_dashes(nx, ny, nz, nmx, nmy, nmz, length, r_angle=rotation_angle))

        for ring in self.aromatic_rings:
            cx = sum(xyz[n][0] for n in ring) / len(ring)
            cy = sum(xyz[n][1] for n in ring) / len(ring)
            cz = sum(xyz[n][2] for n in ring) / len(ring)

            for n, m in zip(ring, ring[1:]):
                nx, ny, nz = xyz[n]
                mx, my, mz = xyz[m]

                aromatic = self._render_3d_aromatic_bond(nx, ny, nz, mx, my, mz, cx, cy, cz)
                if aromatic:
                    veca_x, veca_y, veca_z, vecb_x, vecb_y, vecb_z = aromatic
                    ax, ay, az = nx + veca_x, ny + veca_y, nz + veca_z
                    abx, aby, abz = mx + vecb_x - ax, my + vecb_y - ay, mz + vecb_z - az
                    ab_ln = sqrt(abx ** 2 + aby ** 2 + abz ** 2)
                    if ab_ln < .0001:
                        continue
                    else:
                        xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln))

            i, j = ring[-1], ring[0]
            nx, ny, nz = xyz[i]
            mx, my, mz = xyz[j]
            aromatic = self._render_3d_aromatic_bond(nx, ny, nz, mx, my, mz, cx, cy, cz)
            if aromatic:
                veca_x, veca_y, veca_z, vecb_x, vecb_y, vecb_z = aromatic
                ax, ay, az = nx + veca_x, ny + veca_y, nz + veca_z
                abx, aby, abz = mx + vecb_x - ax, my + vecb_y - ay, mz + vecb_z - az
                ab_ln = sqrt(abx ** 2 + aby ** 2 + abz ** 2)
                if ab_ln < .0001:
                    continue
                else:
                    xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln))

        return ''.join(xml)


class X3domCGR(X3dom):
    __slots__ = ()

    def _render_3d_bonds(self, xyz):
        bonds = self._bonds
        config = self._render_config

        broken = config['broken_color']
        formed = config['formed_color']
        bond_color = config['bond_color']
        bond_radius = config['bond_radius']
        double_space = config['double_space']
        triple_space = config['triple_space']
        r1 = triple_space * sqrt(3) / 3
        r2 = triple_space * sqrt(3) / 6

        xml = []
        lengths = {}
        doubles = {}
        h_triple = triple_space / 2
        ar_bond_colors = defaultdict(dict)
        for n, m, bond in self.bonds():
            order, p_order = bond.order, bond.p_order
            nx, ny, nz = xyz[n]
            mx, my, mz = xyz[m]

            nmx, nmy, nmz = mx - nx, my - ny, mz - nz
            length = sqrt(nmx ** 2 + nmy ** 2 + nmz ** 2)
            if length < .001:
                continue

            rotation_angle = acos(nmy / length)
            lengths[(n, m)] = lengths[(m, n)] = (length, rotation_angle)
            x, y, z = nx + nmx / 2, ny + nmy / 2, nz + nmz / 2
            if order == 1:
                if p_order == 1:
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 4:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'formed_color'
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 2:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{formed}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 3:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")

                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order is None:
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{broken}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                else:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{broken}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.extend(self._render_3d_dashes(nx + dx, ny + dy, nz + dz, nmx, nmy, nmz, length,
                                                      r_angle=rotation_angle, cgr_color='formed_color'))
            elif order == 4:
                if p_order == 4:
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 1:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'broken_color'
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 2:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'broken_color'
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{formed}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 3:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'broken_color'
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")

                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order is None:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'broken_color'
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{broken}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                else:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = None
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{broken}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
            elif order == 2:
                if p_order == 2:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 1:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{broken}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 4:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'formed_color'
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{bond_color}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{broken}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 3:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")

                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order is None:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{broken}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{broken}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                else:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.extend(self._render_3d_dashes(nx + vx, ny + vy, nz + vz, nmx, nmy, nmz, length,
                                                      r_angle=rotation_angle, cgr_color='formed_color'))
            elif order == 3:
                if p_order == 3:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 1:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 4:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'formed_color'
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 2:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{bond_color}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order is None:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                else:
                    third = next((x for x in bonds[n] if x != m), None)
                    if third:
                        ox, oy, oz = xyz[third]
                        nox, noy, noz = ox - nx, oy - ny, oz - nz
                    else:
                        third = next((x for x in bonds[m] if x != n), None)
                        if third:
                            ox, oy, oz = xyz[third]
                            nox, noy, noz = ox - nx, oy - ny, oz - nz
                        else:
                            nox, noy, noz = vector_normal(nmx, nmy, nmz)

                    # normal for plane n m o
                    normx, normy, normz = unit_vector(*plane_normal(nmx, nmy, nmz, nox, noy, noz))

                    # normal for plane n m normal
                    norm_x, norm_y, norm_z = unit_vector(*plane_normal(nmx, nmy, nmz, normx, normy, normz))

                    doubles[n] = doubles[m] = (norm_x, norm_y, norm_z)
                    dx1, dy1, dz1 = norm_x * double_space, norm_y * double_space, norm_z * double_space
                    dx2, dy2, dz2 = normx * double_space, normy * double_space, normz * double_space
                    xml.append(f"    <transform translation='{x + dx1:.2f} {y + dy1:.2f} {z + dz1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x - dx1:.2f} {y - dy1:.2f} {z - dz1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x - dx2:.2f} {y - dy2:.2f} {z - dz2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{broken}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.extend(self._render_3d_dashes(nx + dx2, ny + dy2, nz + dz2, nmx, nmy, nmz, length,
                                                      r_angle=rotation_angle, cgr_color='formed_color'))
            elif order is None:
                if p_order == 1:
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{formed}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 4:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = 'formed_color'
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{formed}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 2:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.append(
                        f"    <transform translation='{x + dx:.2f} {y + dy:.2f} {z + dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{formed}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{formed}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 3:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x + vx:.2f} {y + vy:.2f} {z + vz:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                else:
                    xml.extend(self._render_3d_dashes(nx, ny, nz, nmx, nmy, nmz, length, r_angle=rotation_angle,
                                                      cgr_color='formed_color'))
            else:
                if p_order == 8:
                    xml.extend(self._render_3d_dashes(nx, ny, nz, nmx, nmy, nmz, length, r_angle=rotation_angle))
                elif p_order == 1:
                    dx, dy, dz, doubles = self.__doubles(n, m, nmx, nmy, nmz, xyz, doubles, bonds, double_space)
                    xml.extend(self._render_3d_dashes(nx + dx, ny + dy, nz + dz, nmx, nmy, nmz, length,
                                                   r_angle=rotation_angle, cgr_color='broken_color'))
                    xml.append(
                        f"    <transform translation='{x - dx:.2f} {y - dy:.2f} {z - dz:.2f}' rotation='{nmz:.2f} 0 "
                        f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                        f"          <material diffusecolor='{formed}'>\n          </material>\n"
                        f"       </appearance>\n        <cylinder radius='{bond_radius}' height='{length:.2f}'>\n"
                        "        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 4:
                    ar_bond_colors[n][m] = ar_bond_colors[m][n] = None
                    xml.append(f"    <transform translation='{x:.2f} {y:.2f} {z:.2f}' rotation='{nmz:.2f} 0 "
                               f"{-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n        <appearance>\n"
                               f"          <material diffusecolor='{formed}'>\n          </material>\n"
                               f"       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                elif p_order == 2:
                    vx, vy, vz, x1, y1, z1, x2, y2, z2 = self.__triples(x, y, z, nmx, nmy, nmz,
                                                                        *vector_normal(nmx, nmy, nmz), r1, r2, h_triple)
                    xml.append(f"    <transform translation='{x1:.2f} {y1:.2f} {z1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x2:.2f} {y2:.2f} {z2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.extend(self._render_3d_dashes(nx + vx, ny + vy, nz + vz, nmx, nmy, nmz, length,
                                                      r_angle=rotation_angle, cgr_color='broken_color'))
                elif p_order == 3:
                    third = next((x for x in bonds[n] if x != m), None)
                    if third:
                        ox, oy, oz = xyz[third]
                        nox, noy, noz = ox - nx, oy - ny, oz - nz
                    else:
                        third = next((x for x in bonds[m] if x != n), None)
                        if third:
                            ox, oy, oz = xyz[third]
                            nox, noy, noz = ox - nx, oy - ny, oz - nz
                        else:
                            nox, noy, noz = vector_normal(nmx, nmy, nmz)

                    # normal for plane n m o
                    normx, normy, normz = unit_vector(*plane_normal(nmx, nmy, nmz, nox, noy, noz))

                    # normal for plane n m normal
                    norm_x, norm_y, norm_z = unit_vector(*plane_normal(nmx, nmy, nmz, normx, normy, normz))

                    doubles[n] = doubles[m] = (norm_x, norm_y, norm_z)
                    dx1, dy1, dz1 = norm_x * double_space, norm_y * double_space, norm_z * double_space
                    dx2, dy2, dz2 = normx * double_space, normy * double_space, normz * double_space
                    xml.append(f"    <transform translation='{x + dx1:.2f} {y + dy1:.2f} {z + dz1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x - dx1:.2f} {y - dy1:.2f} {z - dz1:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.append(f"    <transform translation='{x - dx2:.2f} {y - dy2:.2f} {z - dz2:.2f}'"
                               f" rotation='{nmz:.2f} 0 {-nmx:.2f} {rotation_angle:.2f}'>\n      <shape>\n"
                               f"        <appearance>\n          <material diffusecolor='{formed}'>\n"
                               f"          </material>\n       </appearance>\n        <cylinder radius='{bond_radius}'"
                               f" height='{length:.2f}'>\n        </cylinder>\n      </shape>\n    </transform>\n")
                    xml.extend(self._render_3d_dashes(nx + dx2, ny + dy2, nz + dz2, nmx, nmy, nmz, length,
                                                      r_angle=rotation_angle, cgr_color='broken_color'))
                else:
                    xml.extend(self._render_3d_dashes(nx, ny, nz, nmx, nmy, nmz, length, r_angle=rotation_angle,
                                                      cgr_color='broken_color'))

        for ring in self.aromatic_rings:
            cx = sum(xyz[n][0] for n in ring) / len(ring)
            cy = sum(xyz[n][1] for n in ring) / len(ring)
            cz = sum(xyz[n][2] for n in ring) / len(ring)

            for n, m in zip(ring, ring[1:]):
                nx, ny, nz = xyz[n]
                mx, my, mz = xyz[m]

                color = ar_bond_colors[n].get(m)

                aromatic = self._render_3d_aromatic_bond(nx, ny, nz, mx, my, mz, cx, cy, cz)
                if aromatic:
                    veca_x, veca_y, veca_z, vecb_x, vecb_y, vecb_z = aromatic
                    ax, ay, az = nx + veca_x, ny + veca_y, nz + veca_z
                    abx, aby, abz = mx + vecb_x - ax, my + vecb_y - ay, mz + vecb_z - az
                    ab_ln = sqrt(abx ** 2 + aby ** 2 + abz ** 2)
                    if ab_ln < .0001:
                        continue
                    elif color:
                        xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln, cgr_color=color))
                    elif color is None:
                        xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln, r_angle=acos(aby / ab_ln)))
                    else:
                        xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln))

            i, j = ring[-1], ring[0]
            nx, ny, nz = xyz[i]
            mx, my, mz = xyz[j]

            color = ar_bond_colors[i].get(j)
            aromatic = self._render_3d_aromatic_bond(nx, ny, nz, mx, my, mz, cx, cy, cz)
            if aromatic:
                veca_x, veca_y, veca_z, vecb_x, vecb_y, vecb_z = aromatic
                ax, ay, az = nx + veca_x, ny + veca_y, nz + veca_z
                abx, aby, abz = mx + vecb_x - ax, my + vecb_y - ay, mz + vecb_z - az
                ab_ln = sqrt(abx ** 2 + aby ** 2 + abz ** 2)
                if ab_ln < .0001:
                    continue
                elif color:
                    xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln, cgr_color=color))
                elif color is None:
                    xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln, r_angle=acos(aby / ab_ln)))
                else:
                    xml.extend(self._render_3d_dashes(ax, ay, az, abx, aby, abz, ab_ln))
        return ''.join(xml)

    @staticmethod
    def __doubles(i, j, ijx, ijy, ijz, xyz, dbls, bs, ds):
        ix, iy, iz = xyz[i]

        if i in dbls:
            # normal for plane n m o
            norm_x, norm_y, norm_z = plane_normal(ijx, ijy, ijz, *dbls[i])
        elif j in dbls:
            # normal for plane n m o
            norm_x, norm_y, norm_z = plane_normal(ijx, ijy, ijz, *dbls[j])
        else:
            third = next((x for x in bs[i] if x != j), None)
            if third:
                ox, oy, oz = xyz[third]
                nox, noy, noz = ox - ix, oy - iy, oz - iz
            else:
                third = next((x for x in bs[j] if x != i), None)
                if third:
                    ox, oy, oz = xyz[third]
                    nox, noy, noz = ox - ix, oy - iy, oz - iz
                else:
                    nox, noy, noz = vector_normal(ijx, ijy, ijz)

            # normal for plane n m o
            normx, normy, normz = unit_vector(*plane_normal(ijx, ijy, ijz, nox, noy, noz))

            # normal for plane n m normal
            norm_x, norm_y, norm_z = plane_normal(ijx, ijy, ijz, normx, normy, normz)

        dbls[i] = dbls[j] = (norm_x, norm_y, norm_z)
        norm_dist = sqrt(norm_x ** 2 + norm_y ** 2 + norm_z ** 2)

        if norm_dist < .0001:
            coef = ds * 10000
        else:
            coef = ds / norm_dist

        return norm_x * coef, norm_y * coef, norm_z * coef, dbls

    @staticmethod
    def __triples(x, y, z, nmx, nmy, nmz, nox, noy, noz, r1, r2, ht):

        # normal for plane n m o
        normx, normy, normz = unit_vector(*plane_normal(nmx, nmy, nmz, nox, noy, noz))

        # normal for plane n m normal
        norm_x, norm_y, norm_z = unit_vector(*plane_normal(nmx, nmy, nmz, normx, normy, normz))
        vecx, vecy, vecz = norm_x * ht, norm_y * ht, norm_z * ht
        xx, yy, zz = x - normx * r2, y - normy * r2, z - normz * r2
        return normx * r1, normy * r1, normz * r1, xx - vecx, yy - vecy, zz - vecz, xx + vecx, yy + vecy, zz + vecz


__all__ = ['X3domMolecule', 'X3domCGR']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2018 Tagir Akhmetshin <tagirshin@gmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import class_cached_property
from collections import defaultdict
from typing import List
from ..containers import molecule, query  # cyclic imports resolve
from ..exceptions import ValenceError


class Standardize:
    __slots__ = ()
    __class_cache__ = {}

    def canonicalize(self) -> bool:
        """
        Convert molecule to canonical forms of functional groups and aromatic rings without explicit hydrogens.
        """
        s = self.standardize(fix_stereo=False)
        k = self.kekule()
        h = self.implicify_hydrogens(fix_stereo=False)
        t = self.thiele()
        return s or k or h or t

    def standardize(self, *, fix_stereo=True, logging=False) -> bool:
        """
        Standardize functional groups. Return True if any non-canonical group found.

        :param logging: return list of fixed atoms with matched rules.
        """
        neutralized = self.neutralize(fix_stereo=False)
        hs, log = self.__standardize()
        if hs:
            if not neutralized:
                self.flush_cache()
            for n in hs:
                self._calc_implicit(n)
            # second round. need for intersected groups.
            hs, lg = self.__standardize()
            log.extend(lg)
            for n in hs:
                self._calc_implicit(n)
            if fix_stereo:
                self._fix_stereo()
            if logging:
                if neutralized:
                    log.append(((), -1, 'neutralized'))
                return log
            return True
        if neutralized:
            if fix_stereo:
                self._fix_stereo()
            if logging:
                log.append(((), -1, 'neutralized'))
                return log
            return True
        if logging:
            return log
        return False

    def neutralize(self, *, fix_stereo=True) -> bool:
        """
        Transform biradical or dipole resonance structures into neutral form. Return True if structure form changed.
        """
        atoms = self._atoms
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        entries, exits, rads, constrains = self.__entries()
        hs = set()
        while len(rads) > 1:
            n = rads.pop()
            for path in self.__find_delocalize_path(n, rads, constrains):
                l, m, b = path[-1]
                if b == 1:  # required pi-bond
                    continue
                try:
                    atoms[m].valence_rules(charges[m], False, sum(int(y) for x, y in bonds[m].items() if x != l) + b)
                except ValenceError:
                    continue
                self.__patch_path(path)
                radicals[n] = radicals[m] = False
                rads.discard(m)
                hs.add(n)
                hs.update(x for _, x, _ in path)
                break  # path found
            # path not found. atom n keep as is
        while entries and exits:
            n = entries.pop()
            for path in self.__find_delocalize_path(n, exits, constrains):
                l, m, b = path[-1]
                try:
                    atoms[m].valence_rules(charges[m] - 1, radicals[m],
                                           sum(int(y) for x, y in bonds[m].items() if x != l) + b)
                except ValenceError:
                    continue
                self.__patch_path(path)
                charges[n] = charges[m] = 0
                exits.discard(m)
                hs.add(n)
                hs.update(x for _, x, _ in path)
                break  # path from negative atom to positive atom found.
            # path not found. keep negative atom n as is
        if hs:
            self.flush_cache()
            for n in hs:
                self._calc_implicit(n)
                self._calc_hybridization(n)
            if fix_stereo:
                self._fix_stereo()
            return True
        return False

    def implicify_hydrogens(self, *, fix_stereo=True) -> int:
        """
        Remove explicit hydrogen if possible. Works only with Kekule forms of aromatic structures.

        :return: number of removed hydrogens
        """
        atoms = self._atoms
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        explicit = defaultdict(list)
        for n, atom in atoms.items():
            if atom.atomic_number == 1:
                for m in bonds[n]:
                    if atoms[m].atomic_number != 1:
                        explicit[m].append(n)

        to_remove = set()
        for n, hs in explicit.items():
            atom = atoms[n]
            charge = charges[n]
            is_radical = radicals[n]
            len_h = len(hs)
            for i in range(len_h, 0, -1):
                hi = hs[:i]
                explicit_sum = 0
                explicit_dict = defaultdict(int)
                for m, bond in bonds[n].items():
                    if m not in hi:
                        explicit_sum += bond.order
                        explicit_dict[(bond.order, atoms[m].atomic_number)] += 1

                if any(s.issubset(explicit_dict) and all(explicit_dict[k] >= c for k, c in d.items()) and h >= i
                       for s, d, h in atom.valence_rules(charge, is_radical, explicit_sum)):
                    to_remove.update(hi)
                    break
        for n in to_remove:
            self.delete_atom(n)
        if to_remove and fix_stereo:
            self._fix_stereo()
        return len(to_remove)

    def explicify_hydrogens(self, *, fix_stereo=True) -> int:
        """
        Add explicit hydrogens to atoms.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.

        :return: number of added atoms
        """
        to_add = []
        for n, h in self._hydrogens.items():
            try:
                to_add.extend([n] * h)
            except TypeError:
                raise ValenceError(f'atom {{{n}}} has valence error')
        for n in to_add:
            self.add_bond(n, self.add_atom('H'), 1)
        if to_add and fix_stereo:
            self._fix_stereo()
        return len(to_add)

    def check_valence(self) -> List[int]:
        """
        Check valences of all atoms.

        Works only on molecules with aromatic rings in Kekule form.
        :return: list of invalid atoms
        """
        atoms = self._atoms
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        errors = set(atoms)
        for n, atom in atoms.items():
            charge = charges[n]
            is_radical = radicals[n]
            explicit_sum = 0
            explicit_dict = defaultdict(int)
            for m, bond in bonds[n].items():
                order = bond.order
                if order == 4:  # aromatic rings not supported
                    break
                elif order != 8:  # any bond used for complexes
                    explicit_sum += order
                    explicit_dict[(order, atoms[m].atomic_number)] += 1
            else:
                try:
                    rules = atom.valence_rules(charge, is_radical, explicit_sum)
                except ValenceError:
                    pass
                else:
                    for s, d, h in rules:
                        if s.issubset(explicit_dict) and all(explicit_dict[k] >= c for k, c in d.items()):
                            errors.discard(n)
                            break
        return list(errors)

    def clean_isotopes(self) -> bool:
        """
        Clean isotope marks from molecule.
        Return True if any isotope found.
        """
        atoms = self._atoms
        isotopes = [x for x in atoms.values() if x.isotope]
        if isotopes:
            for i in isotopes:
                i._Core__isotope = None
            self.flush_cache()
            self._fix_stereo()
            return True
        return False

    def __standardize(self):
        atom_map = {'charge': self._charges, 'is_radical': self._radicals, 'hybridization': self._hybridizations}
        bonds = self._bonds
        hs = set()
        log = []
        for r, (pattern, atom_fix, bonds_fix) in enumerate(self.__standardize_compiled_rules):
            seen = set()
            for mapping in pattern.get_mapping(self, automorphism_filter=False):
                match = set(mapping.values())
                if not match.isdisjoint(seen):  # skip intersected groups
                    continue
                seen.update(match)
                for n, fix in atom_fix.items():
                    n = mapping[n]
                    for key, value in fix.items():
                        atom_map[key][n] = value
                for n, m, b in bonds_fix:
                    bonds[mapping[n]][mapping[m]]._Bond__order = b
                log.append((tuple(match), r, str(pattern)))
            hs.update(seen)
        return hs, log

    def __patch_path(self, path):
        bonds = self._bonds
        for n, m, b in path:
            bonds[n][m]._Bond__order = b

    def __find_delocalize_path(self, start, finish, constrains):
        bonds = self._bonds
        stack = [(start, n, 0, b.order + 1) for n, b in bonds[start].items() if n in constrains and b.order < 3]
        path = []
        seen = {start}
        while stack:
            last, current, depth, order = stack.pop()
            if len(path) > depth:
                seen.difference_update(x for _, x, _ in path[depth:])
                path = path[:depth]

            path.append((last, current, order))

            if current in finish:
                if depth:  # one bonded ignored. we search double bond transfer! A=A-A >> A-A=A.
                    yield path
                continue  # stop grow

            depth += 1
            seen.add(current)
            diff = -1 if depth % 2 else 1
            stack.extend((current, n, depth, b) for n, b in  # I want 3.8
                         ((n, b.order + diff) for n, b in bonds[current].items() if n not in seen and n in constrains)
                         if 1 <= b <= 3)

    def __entries(self):
        charges = self._charges
        radicals = self._radicals
        atoms = self._atoms
        hybs = self._hybridizations

        transfer = set()
        entries = set()
        exits = set()
        rads = set()
        for n, a in atoms.items():
            if a.atomic_number not in {5, 6, 7, 8, 14, 15, 16, 33, 34, 52} or hybs[n] == 4:
                # filter non-organic set, halogens and aromatics
                continue
            transfer.add(n)
            if charges[n] == -1:
                entries.add(n)
            elif charges[n] == 1:
                exits.add(n)
            elif radicals[n]:
                rads.add(n)
        return entries, exits, rads, transfer

    @class_cached_property
    def __standardize_compiled_rules(self):
        rules = []
        for atoms, bonds, atom_fix, bonds_fix in self.__standardize_rules():
            q = query.QueryContainer()
            for a in atoms:
                q.add_atom(**a)
            for n, m, b in bonds:
                q.add_bond(n, m, b)
            rules.append((q, atom_fix, bonds_fix))
        return rules

    @staticmethod
    def __standardize_rules():
        rules = []

        # Triazine-like
        #
        #   B - N        [B-] - [N+]
        #  /     \       /        \
        # N       B >> [N+]       [B-]
        #  \     /       \        /
        #   B - N        [B-] - [N+]
        #
        atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 1}, {'atom': 'N', 'neighbors': 4, 'hybridization': 1},
                 {'atom': 'B', 'neighbors': 4, 'hybridization': 1}, {'atom': 'N', 'neighbors': 4, 'hybridization': 1},
                 {'atom': 'B', 'neighbors': 4, 'hybridization': 1}, {'atom': 'N', 'neighbors': 4, 'hybridization': 1})
        bonds = ((1, 2, 1), (2, 3, 1), (3, 4, 1), (4, 5, 1), (5, 6, 1), (1, 6, 1))
        atom_fix = {1: {'charge': -1}, 2: {'charge': 1}, 3: {'charge': -1}, 4: {'charge': 1}, 5: {'charge': -1},
                    6: {'charge': 1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Boron
        #
        #     A   A             A      A
        #     |   |             |      |
        # A - B = N - A >> A - [B-] - [N+] - A
        #     |   |             |      |
        #     A   A             A      A
        #
        atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 2}, {'atom': 'N', 'neighbors': 4, 'hybridization': 2})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': -1, 'hybridization': 1}, 2: {'charge': 1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Boron
        #
        #      С            С
        #     //           //
        # B - N   >> B- - [N+]
        #      \            \
        #       A            A
        #
        atoms = ({'atom': 'B'}, {'atom': 'N', 'neighbors': 3, 'hybridization': 2}, {'atom': 'C'})
        bonds = ((1, 2, 1), (2, 3, 2))
        atom_fix = {1: {'charge': -1}, 2: {'charge': 1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Boron
        #
        #     A   A             A      A
        #     |   |             |      |
        # A - B - N - A >> A - [B-] - [N+] - A
        #     |   |             |      |
        #     A   A             A      A
        #
        atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 1}, {'atom': 'N', 'neighbors': 4, 'hybridization': 1})
        bonds = ((1, 2, 1),)
        atom_fix = {1: {'charge': -1}, 2: {'charge': 1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Boron
        #
        #     A    A        A       A
        #     |   //        |      //
        # A - B - N >> A - [B-] - [N+]
        #     |    \        |       \
        #     A     A       A        A
        #
        atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 1}, {'atom': 'N', 'neighbors': 3, 'hybridization': 2})
        bonds = ((1, 2, 1),)
        atom_fix = {1: {'charge': -1}, 2: {'charge': 1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Boron-Nitrogen
        #
        #   A      A       A   A
        #   |      |       |   |
        #  [B-] = [N+] >>  B - N
        #   |      |       |   |
        #   A      A       A   A
        #
        atoms = ({'atom': 'B', 'charge': -1, 'neighbors': 3, 'hybridization': 2},
                 {'atom': 'N', 'charge': 1, 'neighbors': 3, 'hybridization': 2})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': 0, 'hybridization': 1}, 2: {'charge': 0, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Ammonium
        #
        #     A         A
        #     |         |
        #     N    >>  [N+]
        #   / | \     / | \
        #  A  A  A   A  A  A
        #
        atoms = ({'atom': 'N', 'neighbors': 4, 'hybridization': 1},)
        bonds = ()
        atom_fix = {1: {'charge': 1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrone
        #
        #      O          [O-]
        #     //          /
        # C = N  >> C = [N+]
        #      \          \
        #       A          A
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'C'}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrone like
        #
        #      N          [N-]
        #     //          /
        # C = N  >> C = [N+]
        #      \         \
        #       A         A
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2},
                 {'atom': 'C'}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrone
        #
        # N = N = O >> N = [N+] - [O-]
        #     |             |
        #     A             A
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'N'}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrone
        #
        # [N-] - [N+] = O >> N = [N+] - [O-]
        #         |               |
        #         A               A
        #
        atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 3}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'N', 'charge': -1, 'hybridization': 1, 'neighbors': (1, 2)}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 1), (1, 4, 1))
        atom_fix = {2: {'charge': -1, 'hybridization': 1}, 3: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 1), (1, 3, 2))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrone
        #
        # [N+] = N = N - ? >> [N+] = [N+] - [N-] - ?
        #        |                    |
        #        A                    A
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2},
                 {'atom': 'N', 'charge': 1, 'hybridization': 2, 'neighbors': 3}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrone [CN(=O)=N(=O)C]
        #
        # [N+] = N = O >> [N+] = [N+] - O-
        #        |                |
        #        A                A
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'N', 'charge': 1, 'hybridization': 2, 'neighbors': 3}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitro
        #
        #      O          O
        #     //         //
        # C = N  >> C - [N+]
        #      \         \
        #       OH       [O-]
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'C', 'hybridization': 2})
        bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 2))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1}, 4: {'hybridization': 1}}
        bonds_fix = ((1, 4, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitro
        #
        #       [O-]          [O-]
        #       /             /
        # C = [N+ ] >>  C - [N+]
        #      \             \\
        #       OH            O
        #
        atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 3}, {'atom': 'O', 'charge': -1, 'neighbors': 1},
                 {'atom': 'O', 'neighbors': 1}, {'atom': 'C', 'hybridization': 2})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 2))
        atom_fix = {3: {'hybridization': 2}, 4: {'hybridization': 1}}
        bonds_fix = ((1, 3, 2), (1, 4, 1))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitro
        #
        #      O           [O-]
        #     //           /
        # A - N   >> A - [N+]
        #     \\          \\
        #      O           O
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'O', 'neighbors': 1}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitro
        #
        #         O              [O-]
        #        //              /
        # [A-] - N   >> [A-] - [N+]
        #        \\             \\
        #         O              O
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'O', 'neighbors': 1}, {'atom': 'A', 'charge': -1})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitro
        #
        # O : N : O      O = [N+] - [O-]
        #     |      >>       |
        #     A               A
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'O', 'neighbors': 1}, {'atom': 'A'})
        bonds = ((1, 2, 4), (1, 3, 4), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1},
                    3: {'hybridization': 1}}
        bonds_fix = ((1, 2, 1), (1, 3, 2))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # N-same
        #
        #      N           [N-]
        #     //           /
        # A - N   >> A - [N+]
        #     \\          \\
        #      N           N
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'hybridization': 2, 'neighbors': (1, 2)},
                 {'atom': 'N', 'hybridization': 2}, {'atom': 'A'})
        bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrite
        #
        #   O        [O-]
        #  //        /
        # [N-]  >>  N
        #  \\       \\
        #   O        O
        #
        atoms = ({'atom': 'N', 'neighbors': 2, 'charge': -1}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 2), (1, 3, 2))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        #  A - [N-] - [N+] # N  >> A - N = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 2},
                 {'atom': 'N', 'charge': -1, 'neighbors': 2, 'hybridization': 1}, {'atom': 'N', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 3))
        atom_fix = {2: {'charge': 0, 'hybridization': 2}, 3: {'charge': -1, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2), (1, 3, 2))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        #  A - [N+] # N = [N-]  >> A - N = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'charge': 1, 'neighbors': 2},
                 {'atom': 'N', 'charge': -1, 'neighbors': 1}, {'atom': 'A'})
        bonds = ((1, 2, 3), (1, 3, 2), (2, 4, 1))
        atom_fix = {1: {'charge': 1}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        # A - N = N # N >> A - N = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2},
                 {'atom': 'N', 'neighbors': 2, 'hybridization': 2}, {'atom': 'N', 'neighbors': 1})
        bonds = ((1, 2, 2), (1, 3, 3))
        atom_fix = {1: {'charge': 1}, 3: {'charge': -1, 'hybridization': 2}}
        bonds_fix = ((1, 3, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        # A - N = N = N >> A - N = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2},
                 {'atom': 'N', 'neighbors': 2, 'hybridization': 2}, {'atom': 'N', 'neighbors': 1})
        bonds = ((1, 2, 2), (1, 3, 2))
        atom_fix = {1: {'charge': 1}, 3: {'charge': -1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        # A - NH - N # N >> A - N = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2},
                 {'atom': 'N', 'neighbors': 2, 'hybridization': 1}, {'atom': 'N', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 3))
        atom_fix = {1: {'charge': 1}, 2: {'hybridization': 2}, 3: {'charge': -1, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2), (1, 3, 2))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        # [N-] # N = N - A >> [N-] == [N+] == N - A
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1, 'hybridization': 3, 'charge': -1},
                 {'atom': 'N', 'hybridization': 2, 'neighbors': 2})
        bonds = ((1, 2, 3), (1, 3, 2))
        atom_fix = {1: {'charge': 1}, 2: {'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Azide
        #
        # [N-] == N # N >> [N-] == [N+] == [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1, 'charge': -1},
                 {'atom': 'N', 'neighbors': 1})
        bonds = ((1, 2, 2), (1, 3, 3))
        atom_fix = {1: {'charge': 1}, 3: {'hybridization': 2, 'charge': -1}}
        bonds_fix = ((1, 3, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrile oxide
        #
        # - C # N = O >> - C # [N+] - [O-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'O', 'neighbors': 1}, {'atom': 'C'})
        bonds = ((1, 2, 2), (1, 3, 3))
        atom_fix = {1: {'charge': 1}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # N-oxide
        #
        #    |           |
        #  - N - ? >> - [N+] - ?
        #    \\          |
        #     O         [O-]
        #
        atoms = ({'atom': 'N', 'neighbors': (3, 4), 'hybridization': 2}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': 1, 'hybridization': 1}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # N-oxide radical
        #
        #    |         |
        #  - N*  >>  - N
        #    \\        |
        #     O        O*
        #
        atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 2, 'is_radical': True}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'hybridization': 1, 'is_radical': False}, 2: {'is_radical': True, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrous oxide
        #
        # O = N # N >> O = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 3), (1, 3, 2))
        atom_fix = {1: {'charge': 1}, 2: {'charge': -1, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrous oxide
        #
        # [O-] - [N+] # N >> O = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2, 'charge': 1}, {'atom': 'N', 'neighbors': 1},
                 {'atom': 'O', 'neighbors': 1, 'charge': -1, 'hybridization': 1})
        bonds = ((1, 2, 3), (1, 3, 1))
        atom_fix = {2: {'charge': -1, 'hybridization': 2}, 3: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2), (1, 3, 2))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazo
        #
        #  [C-] - [N+] # N  >>  C = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 2}, {'atom': 'N', 'neighbors': 1},
                 {'atom': 'C', 'charge': -1, 'hybridization': 1})
        bonds = ((1, 2, 3), (1, 3, 1))
        atom_fix = {2: {'charge': -1, 'hybridization': 2}, 3: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2), (1, 3, 2))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazo
        #
        # C = N # N >> C = [N+] = [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1}, {'atom': 'C', 'hybridization': 2})
        bonds = ((1, 2, 3), (1, 3, 2))
        atom_fix = {1: {'charge': 1}, 2: {'charge': -1, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazo
        #
        # C # N == N - ? >> [C-] = [N+] = N - ?
        #
        atoms = ({'atom': 'C', 'neighbors': 1, 'hybridization': 3}, {'atom': 'N', 'neighbors': 2, 'hybridization': 3},
                 {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2})
        bonds = ((1, 2, 3), (2, 3, 2))
        atom_fix = {1: {'charge': -1, 'hybridization': 2}, 2: {'charge': 1}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazo
        #
        # A - C # N = N - ? >> A - [C-] = [N+] = N - ?
        #
        atoms = ({'atom': 'C', 'neighbors': 2, 'hybridization': 3}, {'atom': 'N', 'neighbors': 2, 'hybridization': 3},
                 {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2}, {'atom': 'A'})
        bonds = ((1, 2, 3), (2, 3, 2), (1, 4, 1))
        atom_fix = {1: {'charge': -1, 'hybridization': 2}, 2: {'charge': 1}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazo
        #
        # C           C
        #  \           \
        #   N # N >>   [N+] = [N-]
        #  /           /
        # C           C
        #
        atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'neighbors': 1}, {'atom': 'C'}, {'atom': 'C'})
        bonds = ((1, 2, 3), (1, 3, 1), (1, 4, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': -1, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazo
        #
        #    |           |
        #  - N - ? >> - [N+] - ?
        #    \\          |
        #     N         [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': (3, 4), 'hybridization': 2}, {'atom': 'N', 'hybridization': 2})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': 1, 'hybridization': 1}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Diazonium
        #
        #  C - N = [N+]  >>  C - [N+] # N
        #
        atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'charge': 1, 'neighbors': 1},
                 {'atom': 'C', 'hybridization': 1})
        bonds = ((1, 2, 2), (1, 3, 1))
        atom_fix = {1: {'charge': 1, 'hybridization': 3}, 2: {'charge': 0, 'hybridization': 3}}
        bonds_fix = ((1, 2, 3),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Isocyanate
        #
        #  [N+] - [C-] = O  >>  N = C = O
        #
        atoms = ({'atom': 'C', 'charge': -1, 'neighbors': 2},
                 {'atom': 'N', 'charge': 1, 'neighbors': (1, 2), 'hybridization': 1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        atom_fix = {1: {'charge': 0, 'hybridization': 3}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Aromatic N-Oxide
        #
        #  : N :  >>  : [N+] :
        #    \\           \
        #     O           [O-]
        #
        atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 4}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': 1}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Aromatic N-Nitride?
        #
        #  : N :  >>  : [N+] :
        #    \\           \
        #     N           [N-]
        #
        atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 4}, {'atom': 'N', 'hybridization': 2})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': 1}, 2: {'charge': -1, 'hybridization': 1}}
        bonds_fix = ((1, 2, 1),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitroso
        #
        # - [N+] - [O-]  >>  - N = O
        #
        atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 2, 'hybridization': 1},
                 {'atom': 'O', 'charge': -1, 'neighbors': 1})
        bonds = ((1, 2, 1),)
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Iminium
        #
        #  [C+] - N(R2)  >> C = [N+](R2)
        #
        atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 1},
                 {'atom': 'C', 'charge': 1, 'hybridization': 1, 'neighbors': (1, 2, 3)})
        bonds = ((1, 2, 1),)
        atom_fix = {1: {'charge': 1, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Nitrilium
        #
        #  [C+] = N(R)  >>  C # [N+](R)
        #
        atoms = ({'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2},
                 {'atom': 'C', 'charge': 1, 'neighbors': (1, 2), 'hybridization': 2})
        bonds = ((1, 2, 2),)
        atom_fix = {1: {'charge': 1, 'hybridization': 3}, 2: {'charge': 0, 'hybridization': 3}}
        bonds_fix = ((1, 2, 3),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Amide
        #
        #       OH        O
        #      /         //
        # N = C  >> NH - C
        #
        atoms = ({'atom': 'C', 'hybridization': 2}, {'atom': 'O', 'neighbors': 1}, {'atom': 'N', 'hybridization': 2})
        bonds = ((1, 2, 1), (1, 3, 2))
        atom_fix = {2: {'hybridization': 2}, 3: {'hybridization': 1}}
        bonds_fix = ((1, 2, 2), (1, 3, 1))
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Phosphonic
        #
        #       O                   O
        #       |                   |
        #  C - [P+] - [O-]  >>  C - P = O
        #       |                   |
        #       O                   O
        #
        atoms = ({'atom': 'P', 'charge': 1, 'neighbors': 4}, {'atom': 'O', 'charge': -1, 'neighbors': 1},
                 {'atom': 'O'}, {'atom': 'O'}, {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Phosphonium ylide
        #
        #       C                   C
        #       |                   |
        #  C - [P-] - [C+]  >>  C - P = C
        #       |                   |
        #       C                   C
        #
        atoms = ({'atom': 'P', 'charge': -1, 'neighbors': 4}, {'atom': 'C', 'charge': 1, 'neighbors': (1, 2, 3)},
                 {'atom': 'C'}, {'atom': 'C'},
                 {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Hexafluorophosphate
        #
        #   F   F         F   F
        #    \ /           \ /
        # F - P - F >> F - [P-] - F
        #    / \           / \
        #   F   F         F   F
        #
        atoms = ({'atom': 'P', 'neighbors': 6}, {'atom': 'F', 'neighbors': 1}, {'atom': 'F', 'neighbors': 1},
                 {'atom': 'F', 'neighbors': 1}, {'atom': 'F', 'neighbors': 1}, {'atom': 'F', 'neighbors': 1},
                 {'atom': 'F', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1), (1, 6, 1), (1, 7, 1))
        atom_fix = {1: {'charge': -1}}
        bonds_fix = ()
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Sulfoxide
        #
        #       C                   C
        #       |                   |
        #  O = [S+] - [O-]  >>  O = S = O
        #       |                   |
        #       C                   C
        #
        atoms = ({'atom': 'S', 'charge': 1, 'neighbors': 4},
                 {'atom': 'O', 'charge': -1, 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
                 {'atom': 'C'}, {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 1), (1, 5, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Sulfoxonium ylide
        #
        #       C                   C
        #       |                   |
        #  C = [S+] - [O-]  >>  C = S = O
        #       |                   |
        #       C                   C
        #
        atoms = ({'atom': 'S', 'charge': 1, 'neighbors': 4},
                 {'atom': 'O', 'charge': -1, 'neighbors': 1}, {'atom': 'C'}, {'atom': 'C'}, {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 1), (1, 5, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Sulfon
        #
        #           C            C
        #          /            /
        # [O-] - [S+]  >>  O = S
        #          \            \
        #           C            C
        #
        atoms = ({'atom': 'S', 'charge': 1, 'neighbors': 3},
                 {'atom': 'O', 'charge': -1, 'neighbors': 1}, {'atom': 'C'}, {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Sulfonium ylide
        #
        #  C - [S-] - [C+]  >>  C - S = C
        #       |                   |
        #       C                   C
        #
        atoms = ({'atom': 'S', 'charge': -1, 'neighbors': 3},
                 {'atom': 'C', 'charge': 1, 'hybridization': 1, 'neighbors': (1, 2, 3)}, {'atom': 'C'}, {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Sulfine
        #
        #  [O-] - [S+] = C  >>  O = S = C
        #
        atoms = ({'atom': 'S', 'charge': 1, 'neighbors': 2},
                 {'atom': 'O', 'charge': -1, 'neighbors': 1}, {'atom': 'C'})
        bonds = ((1, 2, 1), (1, 3, 2))
        atom_fix = {1: {'charge': 0, 'hybridization': 3}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Silicate Selenite
        #
        #           O             O
        #          /             /
        # [O-] - [Si+]  >>  O = Si
        #          \             \
        #           O             O
        #
        atoms = ({'atom': 'Si', 'charge': 1, 'neighbors': 3}, {'atom': 'O', 'charge': -1, 'neighbors': 1},
                 {'atom': 'O'}, {'atom': 'O'})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        atoms = ({'atom': 'Se', 'charge': 1, 'neighbors': 3}, {'atom': 'O', 'charge': -1, 'neighbors': 1},
                 {'atom': 'O'}, {'atom': 'O'})
        bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1))
        atom_fix = {1: {'charge': 0, 'hybridization': 2}, 2: {'charge': 0, 'hybridization': 2}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Carbon Monoxide
        #
        # [CX1] = O  >> [С-] # [O+]
        #
        atoms = ({'atom': 'C', 'neighbors': 1, 'is_radical': True}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 2), )
        atom_fix = {1: {'charge': -1, 'hybridization': 3, 'is_radical': False}, 2: {'charge': 1, 'hybridization': 3}}
        bonds_fix = ((1, 2, 3),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        # Ozone
        #
        # [O] -- O -- [O]  >>  O == [O+] -- [O-]
        #
        atoms = ({'atom': 'O', 'neighbors': 1, 'is_radical': True}, {'atom': 'O', 'neighbors': 2},
                 {'atom': 'O', 'neighbors': 1, 'is_radical': True})
        bonds = ((1, 2, 1), (2, 3, 1))
        atom_fix = {1: {'hybridization': 2, 'is_radical': False}, 2: {'charge': 1, 'hybridization': 2},
                    3: {'charge': -1, 'is_radical': False}}
        bonds_fix = ((1, 2, 2),)
        rules.append((atoms, bonds, atom_fix, bonds_fix))

        return rules


class StandardizeReaction:
    __slots__ = ()
    __class_cache__ = {}

    def canonicalize(self, fix_mapping: bool = True) -> bool:
        """
        Convert molecules to canonical forms of functional groups and aromatic rings without explicit hydrogens.
        Works only for Molecules.
        Return True if in any molecule found not canonical group.

        :param fix_mapping: Search AAM errors of functional groups.
        """
        total = False
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            if m.canonicalize() and not total:
                total = True

        if fix_mapping and self.fix_mapping():
            return True

        if total:
            self.flush_cache()
        return total

    def standardize(self, fix_mapping: bool = True) -> bool:
        """
        Standardize functional groups. Works only for Molecules.
        Return True if in any molecule found not canonical group.

        :param fix_mapping: Search AAM errors of functional groups.
        """
        total = False
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            if m.standardize() and not total:
                total = True

        if fix_mapping and self.fix_mapping():
            return True

        if total:
            self.flush_cache()
        return total

    def neutralize(self) -> bool:
        """
        Transform biradical or dipole resonance structures into neutral form.
        Works only for Molecules.
        Return True if these groups found in any molecule.
        """
        total = False
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            if m.neutralize() and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def fix_mapping(self) -> bool:
        """
        Fix atom-to-atom mapping of some functional groups. Return True if found AAM errors.
        """
        seen = set()
        if not (self.reactants and self.products):
            return False
        elif not isinstance(self.reactants[0], Standardize):
            raise TypeError('Only Molecules supported')

        for r_pattern, p_pattern, fix in self.__standardize_compiled_rules:
            found = []
            for m in self.reactants:
                for mapping in r_pattern.get_mapping(m, automorphism_filter=False):
                    if mapping[1] not in seen:
                        found.append(({fix.get(k, k): v for k, v in mapping.items()},
                                      {mapping[k]: mapping[v] for k, v in fix.items()}))

            if not found:
                continue
            for m in self.products:
                for mapping in p_pattern.get_mapping(m, automorphism_filter=False):
                    atom = mapping[1]
                    if atom in seen:
                        continue
                    for n, (k, v) in enumerate(found):
                        if k == mapping:
                            break
                    else:
                        continue

                    del found[n]
                    m.remap(v)
                    seen.add(atom)

        if seen:
            self.flush_cache()
            return True
        return False

    def implicify_hydrogens(self) -> int:
        """
        Remove explicit hydrogens if possible

        :return: number of removed hydrogens
        """
        total = 0
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            total += m.implicify_hydrogens()
        if total:
            self.flush_cache()
        return total

    def explicify_hydrogens(self) -> int:
        """
        Add explicit hydrogens to atoms

        :return: number of added atoms
        """
        total = 0
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            total += m.explicify_hydrogens()
        if total:
            self.flush_cache()
        return total

    def thiele(self) -> bool:
        """
        Convert structures to aromatic form. Works only for Molecules.
        Return True if in any molecule found kekule ring
        """
        total = False
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            if m.thiele() and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def kekule(self) -> bool:
        """
        Convert structures to kekule form. Works only for Molecules.
        Return True if in any molecule found aromatic ring
        """
        total = False
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            if m.kekule() and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def clean_isotopes(self) -> bool:
        """
        Clean isotope marks for all molecules in reaction.
        Returns True if in any molecule found isotope.
        """
        flag = False
        for m in self.molecules():
            if not isinstance(m, Standardize):
                raise TypeError('Only Molecules supported')
            if m.clean_isotopes() and not flag:
                flag = True

        if flag:
            self.flush_cache()
        return flag

    def clean_stereo(self):
        """
        Remove stereo data
        """
        for m in self.molecules():
            if not hasattr(m, 'clean_stereo'):
                raise TypeError('Only Molecules and Queries supported')
            m.clean_stereo()
        self.flush_cache()

    def clean2d(self, **kwargs):
        """
        Recalculate 2d coordinates
        """
        for m in self.molecules():
            m.clean2d(**kwargs)
        self.fix_positions()

    def fix_positions(self):
        """
        Fix coordinates of molecules in reaction
        """
        shift_x = 0
        reactants = self.reactants
        amount = len(reactants) - 1
        signs = []
        for m in reactants:
            max_x = self.__fix_positions(m, shift_x)
            if amount:
                max_x += .2
                signs.append(max_x)
                amount -= 1
            shift_x = max_x + 1
        arrow_min = shift_x

        if self.reagents:
            for m in self.reagents:
                max_x = self.__fix_reagent_positions(m, shift_x)
                shift_x = max_x + 1
            if shift_x - arrow_min < 3:
                shift_x = arrow_min + 3
        else:
            shift_x += 3
        arrow_max = shift_x - 1

        products = self.products
        amount = len(products) - 1
        for m in products:
            max_x = self.__fix_positions(m, shift_x)
            if amount:
                max_x += .2
                signs.append(max_x)
                amount -= 1
            shift_x = max_x + 1
        self._arrow = (arrow_min, arrow_max)
        self._signs = tuple(signs)
        self.flush_cache()

    @staticmethod
    def __fix_reagent_positions(molecule, shift_x):
        plane = molecule._plane
        shift_y = .5

        values = plane.values()
        min_x = min(x for x, _ in values) - shift_x
        max_x = max(x for x, _ in values) - min_x
        min_y = min(y for _, y in values) - shift_y
        for n, (x, y) in plane.items():
            plane[n] = (x - min_x, y - min_y)
        return max_x

    @staticmethod
    def __fix_positions(mol, shift_x):
        plane = mol._plane

        left_atom, left_atom_plane = min((x for x in plane.items()), key=lambda x: x[1][0])
        right_atom, right_atom_plane = max((x for x in plane.items()), key=lambda x: x[1][0])

        if len(mol._atoms[left_atom].atomic_symbol) == 2:
            min_x = left_atom_plane[0] - shift_x - .2
        else:
            min_x = left_atom_plane[0] - shift_x

        max_x = right_atom_plane[0]
        max_x -= min_x

        values = plane.values()
        min_y = min(y for _, y in values)
        max_y = max(y for _, y in values)
        mean_y = (max_y + min_y) / 2
        for n, (x, y) in plane.items():
            plane[n] = (x - min_x, y - mean_y)

        r_y = plane[right_atom][1]
        if isinstance(mol, molecule.MoleculeContainer) and -.18 <= r_y <= .18:
            factor = mol._hydrogens[right_atom]
            if factor == 1:
                max_x += .15
            elif factor:
                max_x += .25
        return max_x

    @class_cached_property
    def __standardize_compiled_rules(self):
        rules = []
        for (r_atoms, r_bonds), (p_atoms, p_bonds), fix in self.__standardize_rules():
            r_q = query.QueryContainer()
            p_q = query.QueryContainer()
            for a in r_atoms:
                r_q.add_atom(**a)
            for n, m, b in r_bonds:
                r_q.add_bond(n, m, b)
            for a in p_atoms:
                p_q.add_atom(**a)
            for n, m, b in p_bonds:
                p_q.add_bond(n, m, b)
            rules.append((r_q, p_q, fix))
        return rules

    @staticmethod
    def __standardize_rules():
        rules = []

        # Nitro
        #
        #      O
        #     //
        # * - N+
        #      \
        #       O-
        #
        atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 2, 'charge': 1},
                 {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        fix = {2: 3, 3: 2}
        rules.append(((atoms, bonds), (atoms, bonds), fix))

        # Carbonate
        #
        #      O
        #     //
        # * - C
        #      \
        #       O-
        #
        atoms = ({'atom': 'C', 'neighbors': 3, 'hybridization': 2},
                 {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        fix = {2: 3, 3: 2}
        rules.append(((atoms, bonds), (atoms, bonds), fix))

        # Carbon Acid
        #
        #      O
        #     //
        # * - C
        #      \
        #       OH
        #
        atoms = ({'atom': 'C', 'neighbors': 3, 'hybridization': 2},
                 {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        fix = {2: 3, 3: 2}
        rules.append(((atoms, bonds), (atoms, bonds), fix))

        # Phosphate
        #
        #      *
        #      |
        #  * - P = O
        #      |
        #      OH
        #
        atoms = ({'atom': 'P', 'neighbors': 4, 'hybridization': 2},
                 {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        fix = {2: 3, 3: 2}
        rules.append(((atoms, bonds), (atoms, bonds), fix))

        # Nitro addition
        #
        #      O             O -- *
        #     //            /
        # * - N+   >>  * = N+
        #      \            \
        #       O-           O-
        #
        atoms = ({'atom': 'N', 'neighbors': 3, 'charge': 1, 'hybridization': 2},
                 {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        p_atoms = ({'atom': 'N', 'neighbors': 3, 'charge': 1, 'hybridization': 2},
                   {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 2})
        p_bonds = ((1, 2, 1), (1, 3, 1))
        fix = {2: 3, 3: 2}
        rules.append(((atoms, bonds), (p_atoms, p_bonds), fix))

        # Sulphate addition
        #
        #      O [3]            O -- * [2]
        #     //               /
        # * = S - *   >>  * = S - *
        #     |               \\
        #     O- [2]           O [3]
        #
        atoms = ({'atom': 'S', 'neighbors': 4, 'hybridization': 3},
                 {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1})
        bonds = ((1, 2, 1), (1, 3, 2))
        p_atoms = ({'atom': 'S', 'neighbors': 4, 'hybridization': 3},
                   {'atom': 'O', 'neighbors': 2}, {'atom': 'O', 'neighbors': 1})
        p_bonds = ((1, 2, 1), (1, 3, 2))
        fix = {2: 3, 3: 2}
        rules.append(((atoms, bonds), (p_atoms, p_bonds), fix))

        return rules


__all__ = ['Standardize', 'StandardizeReaction']
